@@ -90,10 +90,40 @@ class FormSshKeypair(SubscriptableBaseModel):
     required:bool=False
     depends_on:str=None
     help_text:str=""
-    
+    local_side:Literal['public','private'] = 'public'
+    """
+    The side of the keypair which belongs to Snowflake. This value will be stored as the secret value, and the other side
+    can be copied/downloaded only during the connection process.
+    """
+    allow_user_provided:bool=True
+    """
+    Allows the user to provide Snowflake's side of the keypair, for cases where one is generated on the outside
+    """
     type:Literal['ssh_keypair'] = 'ssh_keypair'
     secret:bool=True
 
+class FormGpgKeypair(SubscriptableBaseModel):
+    """
+    An GPG Keypair field, which generates public and private GPG keys for asymmetric cryptography.
+    """
+    name:str
+    label:str
+    default_value:str=None
+    required:bool=False
+    depends_on:str=None
+    help_text:str=""
+    local_side:Literal['public','private'] = 'public'
+    """
+    The side of the keypair which belongs to Snowflake. This value will be stored as the secret value, and the other side
+    can be copied/downloaded only during the connection process.
+    """
+    allow_user_provided:bool=True
+    """
+    Allows the user to provide Snowflake's side of the keypair, for cases where one is generated on the outside
+    """    
+    type:Literal['gpg_keypair'] = 'gpg_keypair'
+    secret:bool=True
+
 class FormCheckboxField(SubscriptableBaseModel):
     """
     A field which presents a checkbox
@@ -225,7 +255,7 @@ class FormDropdownField(FormFieldWithDataSource,BaseModel):
 
     multi_select:bool=False
 
-FormFieldBase = Annotated[Union[FormInputField,FormTextAreaField,FormSshKeypair,FormRadioField,FormCheckboxField,FormSliderField,FormDropdownField], Field(discriminator='type')]
+FormFieldBase = Annotated[Union[FormInputField,FormTextAreaField,FormSshKeypair,FormGpgKeypair,FormRadioField,FormCheckboxField,FormSliderField,FormDropdownField], Field(discriminator='type')]
 
 class ConfigurationFormBase(BaseModel,ABC):
     """
@@ -292,15 +322,11 @@ class ConnectionMethod(SubscriptableBaseModel):
     Defines a method of connecting to an application.
     :param str data_source: The name of the connection method, e.g. "OAuth", "API Key", "Credentials"
     :param List[FormFieldBase] fields: A list of fields that are used to collect the connection information from the user.
-    :param List[str] network_addresses: A list of URLs that will be added to a network rule to permit outbound access from Snowflake
-        for the authentication step. Note that for OAuth Authorization flows, it is not necessary to provide the
-        initial URL that the user agent is directed to.
     :param bool oauth: If True, the resulting form will indicate to the user that OAuth will be used. 
         In this scenario, the oauth_parameters function will be called before the connect function.
     """
     name:str
     fields:List[FormFieldBase]
-    network_addresses:List[str]
     oauth:bool=False
 
 
