@@ -5,22 +5,26 @@ Omnata Plugin Runtime.
 Includes data container classes and defines the contract for a plugin.
 """
 from __future__ import annotations
+from decimal import Decimal
 import http
 import pandas
 import jinja2
 from functools import wraps, partial
 from abc import ABC, abstractmethod
-from typing import Callable, Dict, Iterable, List, Literal, Type
+from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Type
 import threading, datetime, json, queue
 from logging import getLogger
 from dateutil.parser import parse
 from jinja2 import Environment
+import pydantic
+from pydantic import BaseModel # pylint: disable=no-name-in-module
 from snowflake.snowpark import Session
-from snowflake.snowpark.functions import col, lit, parse_json, is_null, not_
+from snowflake.snowpark.functions import col, lit
 from snowflake.connector.pandas_tools import write_pandas
 from .forms import ConnectionMethod, OutboundSyncConfigurationForm,InboundSyncConfigurationForm
 from .configuration import STANDARD_OUTBOUND_SYNC_ACTIONS, OutboundSyncAction, OutboundSyncStrategy, OutboundSyncConfigurationParameters, InboundSyncConfigurationParameters, ConnectionConfigurationParameters, StreamConfiguration, SubscriptableBaseModel, SyncConfigurationParameters, StoredStreamConfiguration
 from .rate_limiting import ApiLimits, InterruptedWhileWaitingException, RateLimitState
+from .api import PluginMessage, PluginMessageAbandonedStreams, PluginMessageCancelledStreams, PluginMessageCurrentActivity, PluginMessageStreamProgressUpdate, PluginMessageStreamState, handle_proc_result
 
 logger = getLogger(__name__)
 SortDirectionType = Literal['asc','desc']
@@ -44,6 +48,19 @@ class PluginManifest(SubscriptableBaseModel):
     supports_inbound:bool
     supported_outbound_strategies:List[OutboundSyncStrategy]
 
+class PluginInfo(BaseModel):
+    """
+    Manifest plus other derived information about a plugin which is determined during upload.
+    """
+    manifest: PluginManifest
+    anaconda_packages:List[str]
+    bundled_packages:List[str]
+    icon_source: Optional[str]
+    plugin_class_name: str
+    transformer_class_name: Optional[str]
+    package_source: Literal['function','stage']
+
+
 def jinja_filter(func):
     """
     This annotation designates a function as a jinja filter.
@@ -68,7 +85,8 @@ class SyncRequest(ABC):
                         api_limits:List[ApiLimits],
                         rate_limit_state:Dict[str,RateLimitState],
                         run_deadline:datetime.datetime,
-                        development_mode:bool=False):
+                        development_mode:bool=False,
+                        test_replay_mode:bool=False):
         """
         Constructs a SyncRequest.
 
@@ -101,6 +119,7 @@ class SyncRequest(ABC):
         self._loadbatch_id = 0
         self._loadbatch_id_lock = threading.Lock()
         self.development_mode = development_mode
+        self.test_replay_mode = test_replay_mode
         # This is used internally by the testing framework, when we're loading records in a behave test
         self._prebaked_record_state:pandas.DataFrame = None
         # create a stop requestor to cease thread activity
@@ -144,14 +163,19 @@ class SyncRequest(ABC):
         """
         while not cancellation_token.is_set():
             logger.info("cancel checking worked checking for results")
+            
             with self._snowflake_query_lock:
-                run_cancelled_results = self._session.table('DATA.SYNC_RUN') \
-                    .where((col('SYNC_RUN_ID')==lit(self._run_id))) \
-                    .select(not_(is_null(col('CANCELLED_DATETIME'))).alias('IS_CANCELLED')).collect()
-            if len(run_cancelled_results)==0:
-                raise ValueError(f"Sync run {self._run_id} did not exist when checking for cancellation")
-            if run_cancelled_results[0].IS_CANCELLED:
-                self.apply_cancellation()
+                try:
+                    # this is not ideal, but "Bind variable in stored procedure is not supported yet"
+                    query_result = self._session.sql(f"call {self._source_app_name}.API.PLUGIN_CANCELLATION_CHECK({self._run_id})").collect()
+                    result = json.loads(query_result[0][0])
+                    if result['success'] is not True:
+                        logger.error(f"Error checking cancellation: {result['error']}")
+                    is_cancelled:bool = result['data']['is_cancelled']
+                    if is_cancelled:
+                        self.apply_cancellation()
+                except Exception as e:
+                    logger.error(f"Error updating activity: {e}")
             cancellation_token.wait(10)
         logger.info("cancel checking worker exiting")
     
@@ -226,12 +250,21 @@ class SyncRequest(ABC):
         Avoid lengthy diagnostic messages, anything like this should be logged the normal way.
         """
         logger.info(f"Activity update: {current_activity}")
+        self._plugin_message(PluginMessageCurrentActivity(current_activity=current_activity))
+    
+    def _plugin_message(self,message:PluginMessage):
+        """
+        Sends a message back to the plugin. This is used to send back the results of a sync run.
+        """
+        logger.info(f"Sending plugin message: {message}")
         with self._snowflake_query_lock:
             try:
-                self._session.sql(f"call {self._source_app_name}.API.PLUGIN_ACTIVITY_UPDATE(:1,:2)",
-                              [self._run_id,current_activity]).execute()
+                # this is not ideal, but "Bind variable in stored procedure is not supported yet"
+                handle_proc_result(self._session.sql(f"""call {self._source_app_name}.API.PLUGIN_MESSAGE(
+                                  {self._run_id},
+                                  PARSE_JSON($${json.dumps(message,default=pydantic.json.pydantic_encoder)}$$))""").collect())
             except Exception as e:
-                logger.error(f"Error updating activity: {e}")
+                logger.error(f"Error sending plugin message: {e}",exc_info=True,stack_info=True)
 
 class HttpRateLimiting():
     """
@@ -272,13 +305,16 @@ class OutboundSyncRequest(SyncRequest):
     def __init__(self,run_id:int,
                         session:Session,
                         source_app_name:str,
+                        records_schema_name:str,
+                        records_table_name:str,
                         results_schema_name:str,
                         results_table_name:str,
                         plugin_instance:Type[OmnataPlugin],
                         api_limits:List[ApiLimits],
                         rate_limit_state:Dict[str,RateLimitState],
                         run_deadline:datetime.datetime,
-                        development_mode:bool=False):
+                        development_mode:bool=False,
+                        test_replay_mode:bool=False):
         """
         Constructs an OutboundSyncRequest.
 
@@ -287,9 +323,11 @@ class OutboundSyncRequest(SyncRequest):
         :param OmnataPlugin plugin_instance: The instance of the Omnata Plugin this request is for
         :param ApiLimits api_limits: Constraints to observe when performing HTTP requests
         :param bool development_mode: In development mode, apply_results_queue does not load into Snowflake, instead they are cached locally and can be retrieved via get_queued_results
+        :param bool test_replay_mode: When enabled, it is safe to assume that HTTP requests are hitting a re-recorded log, so there is no need to wait in between polling
         :return: nothing
         """
-        SyncRequest.__init__(self,run_id,session,source_app_name,results_schema_name,results_table_name,plugin_instance,api_limits,rate_limit_state,run_deadline,development_mode)
+        SyncRequest.__init__(self,run_id,session,source_app_name,results_schema_name,results_table_name,plugin_instance,api_limits,rate_limit_state,run_deadline,development_mode,test_replay_mode)
+        self._full_records_table_name = f"{source_app_name}.{records_schema_name}.{records_table_name}"
         self._apply_results:pandas.DataFrame = []
 
     def _get_next_loadbatch_id(self):
@@ -340,8 +378,11 @@ class OutboundSyncRequest(SyncRequest):
         Adds some results to the queue for applying asynchronously
         """
         logger.info(f"Enqueueing {len(results)} results for upload")
+        for required_column in ['IDENTIFIER','RESULT','SUCCESS']:
+            if required_column not in results.columns:
+                raise ValueError(f'{required_column} column was not included in results')
         with self._apply_results_lock:
-            self._apply_results.append(self._preprocess_results_dataframe(results))
+            self._apply_results.append(results)
 
     def get_queued_results(self):
         """
@@ -363,12 +404,8 @@ class OutboundSyncRequest(SyncRequest):
         'LOADBATCH_ID' int
         'RESULT' object
         """
-        for required_column in ['IDENTIFIER','RESULT','SUCCESS']:
-            if required_column not in results_df.columns:
-                raise ValueError(f'{required_column} column was not included in results')
         results_df.set_index('IDENTIFIER',inplace=True, drop=False)
-        append_time = datetime.datetime.utcnow()
-        results_df['APPLY_STATE_DATETIME'] = pandas.Timestamp(append_time).tz_localize('UTC')
+        results_df['APPLY_STATE_DATETIME'] = str(datetime.datetime.now().astimezone())
         if results_df is not None:
             logger.info(f"Applying a queued results dataframe of {len(results_df)} records")
             # change the success flag to an appropriate APPLY STATUS
@@ -377,8 +414,7 @@ class OutboundSyncRequest(SyncRequest):
             results_df = results_df.drop('SUCCESS',axis=1)
             # if results weren't added by enqueue_results, we'll add the status datetime column now
             if 'APPLY_STATE_DATETIME' not in results_df.columns:
-                append_time = datetime.datetime.utcnow()
-                results_df['APPLY_STATE_DATETIME'] = pandas.Timestamp(append_time).tz_localize('UTC')
+                results_df['APPLY_STATE_DATETIME'] = str(datetime.datetime.now().astimezone())
             if 'APP_IDENTIFIER' not in results_df:
                 results_df['APP_IDENTIFIER'] = None
             if 'LOADBATCH_ID' not in results_df:
@@ -394,8 +430,9 @@ class OutboundSyncRequest(SyncRequest):
         # use a random table name with a random string to avoid collisions
         with self._snowflake_query_lock:
             success, nchunks, nrows, _ = write_pandas(conn=self._session._conn._cursor.connection,
-                        df=results_df,
-                        table_name=self._results_table_name,
+                        df=self._preprocess_results_dataframe(results_df),
+                        quote_identifiers=False,
+                        table_name=self._full_results_table_name,
                         auto_create_table=False)
 
     def __dataframe_wrapper(self,data_frame,render_jinja:bool=True):
@@ -458,22 +495,20 @@ class OutboundSyncRequest(SyncRequest):
         if sync_actions is None:
             sync_actions = [action() for action in list(STANDARD_OUTBOUND_SYNC_ACTIONS.values())]
         # ignore null sync actions
-        sync_actions = [s for s in sync_actions if s]
+        sync_action_names:List[str] = [s.action_name for s in sync_actions if s]
         # only used by testing framework when running a behave test
         if self._prebaked_record_state is not None:
             logger.info('returning prebaked record state')
-            dataframe = self._prebaked_record_state[self._prebaked_record_state['SYNC_ACTION'].isin(sync_actions)] # pylint: disable=unsubscriptable-object
-            if len(dataframe)==0:
+            dataframe = self._prebaked_record_state[self._prebaked_record_state['SYNC_ACTION'].isin(sync_action_names)] # pylint: disable=unsubscriptable-object
+            if len(dataframe)==0 or not batched:
                 # no need to do the whole FixedSizeGenerator thing for 0 records
                 return self.__dataframe_wrapper(dataframe,render_jinja)
-            # these were recorded earlier, we should preserve batch size and concurrency from the original
-            #wrapped_df = FixedSizeGenerator(dataframe,self.api_limits.batch_size)
+            # we use map to create an iterable wrapper around the pandas batches which are also iterable
+            # we use an intermediate partial to allow us to pass the extra parameter
             mapfunc = partial(self.__dataframe_wrapper, render_jinja=render_jinja)
             return map(mapfunc, [dataframe])
-            #return self.__dataframe_wrapper(wrapped_df,render_jinja)
-        sync_action_names:List[str] = [action.action_name for action in sync_actions]
         with self._snowflake_query_lock:
-            dataframe = self._session.table(self._full_results_table_name) \
+            dataframe = self._session.table(self._full_records_table_name) \
                     .filter((col("SYNC_ACTION").in_(sync_action_names)) &
                             (col("APPLY_STATE") == lit('ACTIVE'))) \
                     .select(col('IDENTIFIER'),col('SYNC_ACTION'),col('TRANSFORMED_RECORD'))
@@ -505,7 +540,8 @@ class InboundSyncRequest(SyncRequest):
                         rate_limit_state:Dict[str,RateLimitState],
                         run_deadline:datetime.datetime,
                         development_mode:bool = False,
-                        streams:List[StoredStreamConfiguration] = None
+                        streams:List[StoredStreamConfiguration] = None,
+                        test_replay_mode:bool=False
                         ):
         """
         Constructs a record apply request.
@@ -520,9 +556,10 @@ class InboundSyncRequest(SyncRequest):
         :param ApiLimits api_limits: Constraints to observe when performing HTTP requests
         :param bool development_mode: In development mode, apply_results_queue does not load into Snowflake, instead they are cached locally and can be retrieved via get_queued_results
         :param StoredStreamConfiguration streams: The configuration for each stream to fetch
+        :param bool test_replay_mode: When enabled, it is safe to assume that HTTP requests are hitting a re-recorded log, so there is no need to wait in between polling
         :return: nothing
         """
-        SyncRequest.__init__(self,run_id,session,source_app_name,results_schema_name,results_table_name,plugin_instance,api_limits,rate_limit_state,run_deadline,development_mode)
+        SyncRequest.__init__(self,run_id,session,source_app_name,results_schema_name,results_table_name,plugin_instance,api_limits,rate_limit_state,run_deadline,development_mode,test_replay_mode)
         self.streams = streams
         self._streams_dict:Dict[str,StoredStreamConfiguration] = {s.stream_name:s for s in streams}
         self._apply_results:Dict[str,List[pandas.DataFrame]] = {}
@@ -556,15 +593,8 @@ class InboundSyncRequest(SyncRequest):
                     self._apply_results[stream_name] = None
                 self._apply_results = {}
                 # update the inbound stream record counts, so we can see progress
-                # TODO: report back via function call
-                with self._snowflake_query_lock:
-                    stream_record_counts = json.dumps(self._stream_record_counts)
-                    completed_streams = json.dumps(self._completed_streams)
-                    logger.info(f"Updating sync run {self._run_id} stream record counts: {stream_record_counts}, completed streams: {completed_streams}")
-                    self._session.table('DATA.SYNC_RUN').update({
-                        "INBOUND_STREAM_TOTAL_COUNTS": parse_json(lit(stream_record_counts)),
-                        "INBOUND_COMPLETED_STREAMS": parse_json(lit(completed_streams))
-                    },(col('SYNC_RUN_ID')==lit(self._run_id)))
+                self._plugin_message(message=PluginMessageStreamProgressUpdate(stream_total_counts=self._stream_record_counts,
+                                                                               completed_streams=self._completed_streams))
 
     def apply_cancellation(self):
         """
@@ -575,10 +605,7 @@ class InboundSyncRequest(SyncRequest):
         self._thread_cancellation_token.set()
         # any stream which didn't complete at this point is considered cancelled
         cancelled_streams = [stream.stream_name for stream in self.streams if stream.stream_name not in self._completed_streams]
-        with self._snowflake_query_lock:
-            self._session.table('DATA.SYNC_RUN').update({
-                    "INBOUND_CANCELLED_STREAMS": parse_json(lit(json.dumps(cancelled_streams)))
-                },(col('SYNC_RUN_ID')==lit(self._run_id)))
+        self._plugin_message(message=PluginMessageCancelledStreams(cancelled_streams=cancelled_streams))
 
     def apply_deadline_reached(self):
         """
@@ -589,10 +616,7 @@ class InboundSyncRequest(SyncRequest):
         self._thread_cancellation_token.set()
         # any stream which didn't complete at this point is considered abandoned
         abandoned_streams = [stream.stream_name for stream in self.streams if stream.stream_name not in self._completed_streams]
-        with self._snowflake_query_lock:
-            self._session.table('DATA.SYNC_RUN').update({
-                    "INBOUND_ABANDONED_STREAMS": parse_json(lit(json.dumps(abandoned_streams)))
-                },(col('SYNC_RUN_ID')==lit(self._run_id)))
+        self._plugin_message(message=PluginMessageAbandonedStreams(abandoned_streams==abandoned_streams))
 
     def enqueue_results(self,stream_name:str,results:List[Dict],new_state:any):
         """
@@ -683,7 +707,9 @@ class InboundSyncRequest(SyncRequest):
                 results_df['APP_IDENTIFIER'] = results_df['RECORD_DATA'].apply(lambda x: x[primary_key_field])
             else:
                 results_df['APP_IDENTIFIER'] = None
-            results_df['RETRIEVE_DATE'] = datetime.datetime.utcnow()
+            # the timestamps in Snowflake are TIMESTAMP_LTZ, so we upload in string format to ensure the 
+            # timezone information is present.
+            results_df['RETRIEVE_DATE'] = str(datetime.datetime.now().astimezone())
             results_df['STREAM_NAME'] = stream_name
         else:
             results_df = pandas.DataFrame([],columns=['APP_IDENTIFIER','STREAM_NAME','RECORD_DATA','RETRIEVE_DATE'])
@@ -703,7 +729,6 @@ class InboundSyncRequest(SyncRequest):
                         table_name=self._full_results_table_name,
                         quote_identifiers=False, # already done in get_temp_table_name
                         #schema='INBOUND_RAW', # it seems to be ok to provide schema in the table name
-                        auto_create_table=True,
                         table_type='transient')
                 if not success:
                     raise ValueError(f"Failed to write results to Snowflake table {self._full_results_table_name}")
@@ -723,13 +748,7 @@ class InboundSyncRequest(SyncRequest):
         Updates the SYNC table to have the latest stream states.
         TODO: This should be done in concert with the results, revisit
         """
-        with self._snowflake_query_lock:
-            try:
-                self._session.sql(f"call {self._source_app_name}.API.PLUGIN_STREAM_STATE_UPDATE(:1,PARSE_JSON($$:2$$))",
-                              [self._run_id,json.dumps(self._latest_states)]).execute()
-            except Exception as e:
-                logger.error(f"Error updating activity: {e}")
-
+        self._plugin_message(PluginMessageStreamState(stream_state=self._latest_states))
 
 class OAuthParameters(SubscriptableBaseModel):
     """
@@ -757,6 +776,32 @@ class ConnectResponse(SubscriptableBaseModel):
     connection_secrets:dict=None
     network_addresses:List[str]=None
 
+class BillingEvent(BaseModel):
+    """
+    Represents a Snowflake billing event.
+    The description of fields can be found in the Snowflake product docs for custom event billing for Native Applications
+    """
+    billing_class: str
+    billing_subclass: Optional[str]
+    start_timestamp: Optional[datetime.datetime] = None
+    base_charge: Decimal
+    objects:List[str] = []
+    additional_info: Dict[str,Any] = {}
+
+class BillingEventRequest(BaseModel):
+    """
+    Represents a request to provide billing events for that day.
+    The Omnata engine provides the number of runs for the most frequent inbound and outbound syncs, and the sync ids
+    """
+    billing_schedule: Literal['DAILY']
+    inbound_most_frequent_run_count: int
+    inbound_most_frequent_sync_id: int
+    outbound_most_frequent_run_count: int
+    outbound_most_frequent_sync_id: int
+
+#BillingEventRequest = Annotated[Union[DailyBillingEventRequest,...],Field(discriminator='billing_schedule')]
+
+
 class OmnataPlugin(ABC):
     """
     Class which defines the contract for an Omnata Push Plugin
@@ -783,6 +828,22 @@ class OmnataPlugin(ABC):
         :rtype List[ConnectionMethod]
         """
         raise NotImplementedError('Your plugin class must implement the connection_form method')
+    
+    @abstractmethod
+    def network_addresses(self, parameters:ConnectionConfigurationParameters) -> List[str]:
+        """
+        Returns a list of network addresses that are required to connect to the app.
+        This will be called after the connection form is completed, so that collected information can be used to build the list.
+        Note that at this point, no external access is possible.
+
+        :param ConnectionConfigurationParameters parameters the parameters of the connection, configured so far.
+        :return A list of domains that will be added to a network rule to permit outbound access from Snowflake
+        for the authentication step. Note that for OAuth Authorization flows, it is not necessary to provide the
+        initial URL that the user agent is directed to.
+        :rtype List[str]
+        """
+        raise NotImplementedError('Your plugin class must implement the network_addresses method')
+
 
     def outbound_configuration_form(self,parameters:OutboundSyncConfigurationParameters) -> OutboundSyncConfigurationForm:
         """
@@ -862,6 +923,14 @@ class OmnataPlugin(ABC):
         Defines the API limits in place for the app's API
         """
         return []
+    
+    def create_billing_events(self,request: BillingEventRequest) -> List[BillingEvent]:
+        """
+        Creates billing events for the day, these will be submitted to the Snowflake event billing API.
+        Note that the Snowflake API is strictly rate limited, so only a very small number of events
+        should be returned.
+        """
+        return []
 
     def additional_loggers(self) -> List[str]:
         """
@@ -999,7 +1068,7 @@ def managed_outbound_processing(concurrency:int, batch_size:int):
             #    raise ValueError('To use the managed_outbound_processing decorator, API constraints must be defined. These can be provided in the response to the connect method')
             logger.info(f"Batch size: {batch_size}. Concurrency: {concurrency}")
             if len(method_args)==0:
-                raise ValueError('You must provide at least one method argument, and the first argument must be a DataFrame or DataFrame generator (from outbound_sync_request.get_records_to_*)')
+                raise ValueError('You must provide at least one method argument, and the first argument must be a DataFrame or DataFrame generator (from outbound_sync_request.get_records)')
             first_arg = method_args[0]
             logger.info(first_arg.__class__.__name__)
             if first_arg.__class__.__name__ == 'DataFrame':
@@ -1007,7 +1076,7 @@ def managed_outbound_processing(concurrency:int, batch_size:int):
             elif hasattr(first_arg,'__next__'):
                 logger.info('managed_outbound_processing received an iterator function')
             else:
-                raise ValueError(f'The first argument to a @managed_outbound_processing method must be a DataFrame or DataFrame generator (from outbound_sync_request.get_records_to_*). Instead, a {first_arg.__class__.__name__} was provided.')
+                raise ValueError(f'The first argument to a @managed_outbound_processing method must be a DataFrame or DataFrame generator (from outbound_sync_request.get_records). Instead, a {first_arg.__class__.__name__} was provided.')
 
             # put the record iterator on the queue, ready for the first task to read it
             fixed_size_generator = FixedSizeGenerator(first_arg,batch_size = batch_size)
