@@ -4,19 +4,20 @@ import sys
 import os
 import importlib
 import datetime
-from logging import getLogger
-from typing import Dict, List, Union, Optional
-from omnata_plugin_runtime.forms import ConnectionMethod
-from omnata_plugin_runtime.logging import OmnataPluginLogHandler
-from omnata_plugin_runtime.rate_limiting import ApiLimits
+import logging
+import time
+from typing import Dict, List, Optional
 from pydantic import BaseModel, parse_obj_as # pylint: disable=no-name-in-module
 from pydantic.json import pydantic_encoder as pydantic_json_encoder # pylint:disable=no-name-in-module
 from snowflake.snowpark import Session
-from omnata_plugin_runtime.omnata_plugin import OmnataPlugin, HttpRateLimiting, OutboundSyncRequest, InboundSyncRequest
-from omnata_plugin_runtime.configuration import OutboundSyncConfigurationParameters,SyncDirection,InboundSyncConfigurationParameters, OutboundSyncStrategy, StoredConfigurationValue,SyncConfigurationParameters,InboundSyncStreamsConfiguration, StreamConfiguration, StoredStreamConfiguration, SyncConfigurationParameters, StoredMappingValue
-from omnata_plugin_runtime.rate_limiting import ApiLimits, RetryLaterException, RateLimitState, RequestRateLimit
-from omnata_plugin_runtime.api import ApplyPayload
-logger = getLogger(__name__)
+from .forms import ConnectionMethod
+from .logging import OmnataPluginLogHandler
+from .omnata_plugin import BillingEvent,BillingEventRequest, OmnataPlugin, HttpRateLimiting, OutboundSyncRequest, InboundSyncRequest
+from .configuration import OutboundSyncConfigurationParameters,InboundSyncConfigurationParameters, OutboundSyncStrategy, StoredConfigurationValue,SyncConfigurationParameters,StreamConfiguration, StoredStreamConfiguration
+from .rate_limiting import ApiLimits, RateLimitState
+from .api import ApplyPayload, handle_proc_result
+# set the logger class to our custom logger so that pydantic errors are handled correctly
+logger = logging.getLogger(__name__)
 # this is the new API for secrets access (https://docs.snowflake.com/en/LIMITEDACCESS/secret-api-reference)
 import _snowflake # pylint: disable=import-error
 IMPORT_DIRECTORY_NAME = "snowflake_import_directory"
@@ -50,7 +51,7 @@ class PluginEntrypoint():
                                                 sync_run_id=request.run_id)
         omnata_log_handler.register(request.logging_level,self._plugin_instance.additional_loggers())
         # construct some generic parameters for the purpose of getting the api limits
-        base_parameters = SyncConfigurationParameters(request.connection_method,
+        base_parameters = SyncConfigurationParameters(connection_method=request.connection_method,
                             connection_parameters=request.connection_parameters,
                             connection_secrets=connection_secrets,
                             sync_parameters=request.sync_parameters,
@@ -65,6 +66,7 @@ class PluginEntrypoint():
             if api_limit.endpoint_category not in request.rate_limits_state:
                 request.rate_limits_state[api_limit.endpoint_category] = RateLimitState()
         logger.info(f"Rate limits state: {json.dumps(request.rate_limits_state, default=pydantic_json_encoder)}")
+        return_dict={}
         if request.sync_direction=='outbound':
             parameters = OutboundSyncConfigurationParameters(
                 connection_method=request.connection_method,
@@ -79,20 +81,20 @@ class PluginEntrypoint():
             outbound_sync_request = OutboundSyncRequest(run_id=request.run_id,
                                             session=self._session,
                                             source_app_name=request.source_app_name,
+                                            records_schema_name=request.records_schema_name,
+                                            records_table_name=request.records_table_name,
                                             results_schema_name=request.results_schema_name,
                                             results_table_name=request.results_table_name,
                                             plugin_instance=self._plugin_instance,
-                                            api_limits=request.api_limits,
+                                            api_limits=request.api_limit_overrides,
                                             rate_limit_state=request.rate_limits_state,
                                             run_deadline=datetime.datetime.now() + datetime.timedelta(hours=4),
                                             development_mode=False
                                             )
             
             with HttpRateLimiting(outbound_sync_request,parameters):
-                combined_results = self._plugin_instance.sync_outbound(parameters,outbound_sync_request)
-                if combined_results is not None:
-                    outbound_sync_request.enqueue_results(combined_results)
-                    outbound_sync_request.apply_results_queue()
+                self._plugin_instance.sync_outbound(parameters,outbound_sync_request)
+                outbound_sync_request.apply_results_queue()
         
         elif request.sync_direction=='inbound':
             logger.info("Running inbound sync")
@@ -134,7 +136,7 @@ class PluginEntrypoint():
             
             for stream in streams_list:
                 if stream.stream_name in request.latest_stream_state:
-                    stream.latest_state = request.request.latest_stream_state[stream.stream_name]
+                    stream.latest_state = request.latest_stream_state[stream.stream_name]
                     logger.info(f"Updating stream state for {stream.stream_name}: {stream.latest_state}")
                 else:
                     logger.info(f"Existing stream state for {stream.stream_name} not found")
@@ -150,7 +152,7 @@ class PluginEntrypoint():
                                             results_schema_name=request.results_schema_name,
                                             results_table_name=request.results_table_name,
                                             plugin_instance=self._plugin_instance,
-                                            api_limits=request.api_limits,
+                                            api_limits=request.api_limit_overrides,
                                             rate_limit_state=request.rate_limits_state,
                                             run_deadline=datetime.datetime.now() + datetime.timedelta(hours=4),
                                             development_mode=False,
@@ -161,20 +163,20 @@ class PluginEntrypoint():
             logger.info(f"inbound sync request: {inbound_sync_request}")
             #plugin_instance._inbound_sync_request = outbound_sync_request
             with HttpRateLimiting(inbound_sync_request,parameters):
-                combined_results = self._plugin_instance.sync_inbound(parameters,inbound_sync_request)
+                self._plugin_instance.sync_inbound(parameters,inbound_sync_request)
             logger.info(f"Finished invoking plugin")
             inbound_sync_request.update_activity(f"Staging remaining records")
             logger.info(f"Calling apply_results_queue")
             inbound_sync_request.apply_results_queue()
             omnata_log_handler.flush()
+            return_dict["streams"] = [s.dict() for s in streams_list]
             # we need to calculate counts for:
             # CHANGED_COUNT by counting up the records in INBOUND_STREAM_RECORD_COUNTS
         logger.info('Finished applying records')
         omnata_log_handler.flush()
         omnata_log_handler.unregister()
-        return {
-            "streams": dict(streams_list)
-        }
+        
+        return return_dict
 
     def configuration_form(self,
                            connection_method:str,
@@ -222,16 +224,36 @@ class PluginEntrypoint():
         return script_result
         
     def connection_form(self):
-        try:
-            logger.info('Entered connection_form method')
-            form:List[ConnectionMethod] = self._plugin_instance.connection_form()
-            return [f.dict() for f in form]
-        except Exception as exception:
-            logger.error(str(exception),exc_info=True,stack_info=True)
-            return {
-                "success": False,
-                "error": str(exception)
-            }
+        logger.info('Entered connection_form method')
+        form:List[ConnectionMethod] = self._plugin_instance.connection_form()
+        return [f.dict() for f in form]
+
+    def create_billing_events(self,session, event_request:Dict):
+        logger.info('Entered create_billing_events method')
+        request = parse_obj_as(BillingEventRequest,event_request)
+        events:List[BillingEvent] = self._plugin_instance.create_billing_events(request)
+        # create each billing event, waiting a minute between each one
+        first_time = True
+        for billing_event in events:
+            if not first_time:
+                time.sleep(60)
+            else:
+                first_time = False
+            current_time = int(time.time() *1000)
+            billing_event_time = int(billing_event.start_timestamp.replace(tzinfo=datetime.timezone.utc).timestamp() * 1000)
+            event_query = f"""call SYSTEM$CREATE_BILLING_EVENT(
+                    $${billing_event.billing_class}$$,
+                    $${billing_event.billing_subclass}$$,
+                    {billing_event_time},
+                    {current_time},
+                    {str(billing_event.base_charge)},
+                    $${json.dumps(billing_event.objects)}$$,
+                    $${json.dumps(billing_event.additional_info)}$$)
+                    """
+            logger.info(f'Executing billing event query: {event_query}')
+            handle_proc_result(session.sql(event_query).collect())
+
+        return [e.dict() for e in events]
 
     def get_secrets(self,oauth_secret_name:Optional[str],other_secrets_name:Optional[str]) -> Dict[str,StoredConfigurationValue]:
         connection_secrets={}
@@ -244,9 +266,17 @@ class PluginEntrypoint():
             except Exception as exception:
                 logger.error(f'Error parsing secrets content: {str(exception)}')
                 raise ValueError(f'Error parsing secrets content: {str(exception)}')
-            connection_secrets = parse_obj_as(Dict[str,StoredConfigurationValue],other_secrets)
+            connection_secrets = {**connection_secrets,**parse_obj_as(Dict[str,StoredConfigurationValue],other_secrets)}
         return connection_secrets
     
+    def network_addresses(self,method:str,connection_parameters:Dict) -> List[str]:
+        logger.info('Entered network_addresses method')
+        logger.info(f'Connection parameters: {connection_parameters}')
+        from omnata_plugin_runtime.omnata_plugin import ConnectionConfigurationParameters
+        return self._plugin_instance.network_addresses(ConnectionConfigurationParameters(connection_method=method,
+                                                                connection_parameters=parse_obj_as(Dict[str,StoredConfigurationValue],connection_parameters),
+                                                                connection_secrets={}))
+    
     def connect(self,method,connection_parameters:Dict,network_rule_name:str,oauth_secret_name:Optional[str],other_secrets_name:Optional[str]):
         logger.info('Entered connect method')
         logger.info(f'Connection parameters: {connection_parameters}')
